import argparse
import asyncio
import base64
import io
import json
import os
import random
import time
import warnings
from dataclasses import dataclass
from datetime import datetime
from typing import Any, AsyncGenerator, Collection, Dict, List, Optional, Tuple

import numpy as np
from .backend_request_func import (ASYNC_REQUEST_FUNCS, RequestFuncInput,
                                  RequestFuncOutput)
from datasets import load_dataset
from PIL.Image import Image
from tqdm.asyncio import tqdm
from transformers import PreTrainedTokenizerBase

try:
    from vllm.transformers_utils.tokenizer import get_tokenizer
except ImportError:
    from .backend_request_func import get_tokenizer

try:
    from vllm.utils import FlexibleArgumentParser
except ImportError:
    from argparse import ArgumentParser as FlexibleArgumentParser

MILLISECONDS_TO_SECONDS_CONVERSION = 1000



class LoadTestingOperator:
    def __init__(self):
        # Initialize any necessary attributes or connections here
        pass

    def load_test(self, args: argparse.Namespace):
        """
        Start a load test with the given parameters.
        
        :param test_plan: Path to the test plan file
        :param duration: Duration of the load test in seconds
        :param concurrency: Number of concurrent users
        """
        print(args)
        random.seed(args.seed)
        np.random.seed(args.seed)

        backend = args.backend
        model_id = args.model
        tokenizer_id = args.tokenizer if args.tokenizer is not None else args.model

        if args.base_url is not None:
            api_url = f"{args.base_url}{args.endpoint}"
            base_url = f"{args.base_url}"
        else:
            api_url = f"http://{args.host}:{args.port}{args.endpoint}"
            base_url = f"http://{args.host}:{args.port}"

        tokenizer = get_tokenizer(tokenizer_id,
                                trust_remote_code=args.trust_remote_code)

        if args.dataset is not None:
            warnings.warn(
                "The '--dataset' argument will be deprecated in the next "
                "release. Please use '--dataset-name' and "
                "'--dataset-path' in the future runs.",
                stacklevel=2)
            input_requests = sample_sharegpt_requests(
                dataset_path=args.dataset,
                num_requests=args.num_prompts,
                tokenizer=tokenizer,
                fixed_output_len=args.sharegpt_output_len,
            )

        elif args.dataset_name == "sharegpt":
            input_requests = sample_sharegpt_requests(
                dataset_path=args.dataset_path,
                num_requests=args.num_prompts,
                tokenizer=tokenizer,
                fixed_output_len=args.sharegpt_output_len,
            )

        elif args.dataset_name == "sonnet":
            # Do not format the prompt, pass to message directly
            if args.backend == "openai-chat":
                input_requests = sample_sonnet_requests(
                    dataset_path=args.dataset_path,
                    num_requests=args.num_prompts,
                    input_len=args.sonnet_input_len,
                    output_len=args.sonnet_output_len,
                    prefix_len=args.sonnet_prefix_len,
                    tokenizer=tokenizer,
                )
                input_requests = [(prompt, prompt_len, output_len, None)
                                for prompt, prompt_formatted, prompt_len,
                                output_len, _ in input_requests]
            else:
                assert (
                    tokenizer.chat_template or tokenizer.default_chat_template
                ), "Tokenizer/model must have chat template for sonnet dataset."
                input_requests = sample_sonnet_requests(
                    dataset_path=args.dataset_path,
                    num_requests=args.num_prompts,
                    input_len=args.sonnet_input_len,
                    output_len=args.sonnet_output_len,
                    prefix_len=args.sonnet_prefix_len,
                    tokenizer=tokenizer,
                )
                input_requests = [(prompt_formatted, prompt_len, output_len, None)
                                for prompt, prompt_formatted, prompt_len,
                                output_len, _ in input_requests]

        elif args.dataset_name == "hf":
            input_requests = sample_hf_requests(
                dataset_path=args.dataset_path,
                dataset_subset=args.hf_subset,
                dataset_split=args.hf_split,
                num_requests=args.num_prompts,
                tokenizer=tokenizer,
                random_seed=args.seed,
                fixed_output_len=args.hf_output_len,
            )

        elif args.dataset_name == "random":
            input_requests = sample_random_requests(
                prefix_len=args.random_prefix_len,
                input_len=args.random_input_len,
                output_len=args.random_output_len,
                num_prompts=args.num_prompts,
                range_ratio=args.random_range_ratio,
                tokenizer=tokenizer,
            )

        else:
            raise ValueError(f"Unknown dataset: {args.dataset_name}")

        gootput_config_dict = check_goodput_args(args)

        benchmark_result = asyncio.run(
            benchmark(
                backend=backend,
                api_url=api_url,
                base_url=base_url,
                model_id=model_id,
                tokenizer=tokenizer,
                input_requests=input_requests,
                logprobs=args.logprobs,
                best_of=args.best_of,
                request_rate=args.request_rate,
                burstiness=args.burstiness,
                disable_tqdm=args.disable_tqdm,
                profile=args.profile,
                selected_percentile_metrics=args.percentile_metrics.split(","),
                selected_percentiles=[
                    float(p) for p in args.metric_percentiles.split(",")
                ],
                ignore_eos=args.ignore_eos,
                gootput_config_dict=gootput_config_dict,
                max_concurrency=args.max_concurrency,
            ))

        # Save config and results to json
        if args.save_result:
            result_json: Dict[str, Any] = {}

            # Setup
            current_dt = datetime.now().strftime("%Y%m%d-%H%M%S")
            result_json["date"] = current_dt
            result_json["backend"] = backend
            result_json["model_id"] = model_id
            result_json["tokenizer_id"] = tokenizer_id
            result_json["best_of"] = args.best_of
            result_json["num_prompts"] = args.num_prompts

            # Metadata
            if args.metadata:
                for item in args.metadata:
                    if "=" in item:
                        kvstring = item.split("=")
                        result_json[kvstring[0].strip()] = kvstring[1].strip()
                    else:
                        raise ValueError(
                            "Invalid metadata format. Please use KEY=VALUE format."
                        )

            # Traffic
            result_json["request_rate"] = (
                args.request_rate if args.request_rate < float("inf") else "inf")
            result_json["burstiness"] = args.burstiness
            result_json["max_concurrency"] = args.max_concurrency

            # Merge with benchmark result
            result_json = {**result_json, **benchmark_result}

            # Save to file
            base_model_id = model_id.split("/")[-1]
            max_concurrency_str = (f"-concurrency{args.max_concurrency}"
                                if args.max_concurrency is not None else "")
            file_name = f"{backend}-{args.request_rate}qps{max_concurrency_str}-{base_model_id}-{current_dt}.json"  #noqa
            if args.result_filename:
                file_name = args.result_filename
            if args.result_dir:
                file_name = os.path.join(args.result_dir, file_name)
            with open(file_name, "w", encoding='utf-8') as outfile:
                json.dump(result_json, outfile)

        def stop_test(self, test_id):
            """
            Stop a load test with the given test ID.
            
            :param test_id: ID of the test to stop
            """
            # Implement the logic to stop the load test
            print(f"Stopping load test with ID {test_id}.")
            # Add your load testing logic here

        def get_test_status(self, test_id):
            """
            Get the status of a load test with the given test ID.
            
            :param test_id: ID of the test to get status
            """
            # Implement the logic to get the status of the load test
            print(f"Getting status for load test with ID {test_id}.")
            # Add your load testing logic here




@dataclass
class BenchmarkMetrics:
    completed: int
    total_input: int
    total_output: int
    request_throughput: float
    request_goodput: float
    output_throughput: float
    total_token_throughput: float
    mean_input_token_len: int
    median_input_token_len: int
    std_input_token_len: int
    percentiles_input_token_len: List[Tuple[float, int]]
    mean_output_token_len: int
    median_output_token_len: int
    std_output_token_len: int
    percentiles_output_token_len: List[Tuple[float, int]]
    mean_fs_token_len: int
    median_fs_token_len: int
    std_fs_token_len: int
    percentiles_fs_token_len: List[Tuple[float, int]]
    mean_input_token_len: int
    median_input_token_len: int
    std_input_token_len: int
    percentiles_input_token_len: List[Tuple[float, int]]
    mean_ttft_ms: float
    median_ttft_ms: float
    std_ttft_ms: float
    percentiles_ttft_ms: List[Tuple[float, float]]
    mean_ttfs_ms: float
    median_ttfs_ms: float
    std_ttfs_ms: float
    percentiles_ttfs_ms: List[Tuple[float, float]]
    mean_tpot_ms: float
    median_tpot_ms: float
    std_tpot_ms: float
    percentiles_tpot_ms: List[Tuple[float, float]]
    mean_itl_ms: float
    median_itl_ms: float
    std_itl_ms: float
    percentiles_itl_ms: List[Tuple[float, float]]
    # E2EL stands for end-to-end latency per request.
    # It is the time taken on the client side from sending
    # a request to receiving a complete response.
    mean_e2el_ms: float
    median_e2el_ms: float
    std_e2el_ms: float
    percentiles_e2el_ms: List[Tuple[float, float]]


def sample_sharegpt_requests(
    dataset_path: str,
    num_requests: int,
    tokenizer: PreTrainedTokenizerBase,
    fixed_output_len: Optional[int] = None,
) -> List[Tuple[str, int, int, None]]:
    # Load the dataset.
    with open(dataset_path, encoding='utf-8') as f:
        dataset = json.load(f)
    # Filter out the conversations with less than 2 turns.
    dataset = [data for data in dataset if len(data["conversations"]) >= 2]
    # Only keep the first two turns of each conversation.
    dataset = [(data["conversations"][0]["value"],
                data["conversations"][1]["value"]) for data in dataset]

    # Shuffle the dataset.
    random.shuffle(dataset)

    # Filter out sequences that are too long or too short
    filtered_dataset: List[Tuple[str, int, int]] = []
    for i in range(len(dataset)):
        if len(filtered_dataset) == num_requests:
            break

        # Tokenize the prompts and completions.
        prompt = dataset[i][0]
        prompt_token_ids = tokenizer(prompt).input_ids
        completion = dataset[i][1]
        completion_token_ids = tokenizer(completion).input_ids
        prompt_len = len(prompt_token_ids)
        output_len = len(completion_token_ids
                         ) if fixed_output_len is None else fixed_output_len
        if prompt_len < 4 or (fixed_output_len is None and output_len < 4):
            # Prune too short sequences.
            continue
        if prompt_len > 1024 or prompt_len + output_len > 2048:
            # Prune too long sequences.
            continue
        filtered_dataset.append((prompt, prompt_len, output_len, None))

    return filtered_dataset


def sample_sonnet_requests(
    dataset_path: str,
    num_requests: int,
    input_len: int,
    output_len: int,
    prefix_len: int,
    tokenizer: PreTrainedTokenizerBase,
) -> List[Tuple[str, str, int, int, None]]:
    assert (
        input_len > prefix_len
    ), "'args.sonnet-input-len' must be greater than 'args.prefix-input-len'."

    # Load the dataset.
    with open(dataset_path, encoding='utf-8') as f:
        poem_lines = f.readlines()

    # Tokenize the poem lines.
    poem_token_ids = tokenizer(poem_lines).input_ids
    average_poem_len = sum(
        len(token_ids) for token_ids in poem_token_ids) / len(poem_token_ids)

    # Base prefix for all requests.
    base_prompt = "Pick as many lines as you can from these poem lines:\n"
    base_message = [{
        "role": "user",
        "content": base_prompt,
    }]
    base_prompt_formatted = tokenizer.apply_chat_template(
        base_message, add_generation_prompt=True, tokenize=False)
    base_prompt_offset = len(tokenizer(base_prompt_formatted).input_ids)

    assert (
        input_len > base_prompt_offset
    ), f"Please set 'args.sonnet-input-len' higher than {base_prompt_offset}."
    num_input_lines = round(
        (input_len - base_prompt_offset) / average_poem_len)

    # First approximately `prefix_len` number of tokens in the
    # prompt are fixed poem lines.
    assert (
        prefix_len > base_prompt_offset
    ), f"Please set 'args.sonnet-prefix-len' higher than {base_prompt_offset}."

    num_prefix_lines = round(
        (prefix_len - base_prompt_offset) / average_poem_len)
    prefix_lines = poem_lines[:num_prefix_lines]

    # Sample the rest of lines per request.
    sampled_requests: List[Tuple[str, int, int]] = []
    for _ in range(num_requests):
        num_lines_needed = num_input_lines - num_prefix_lines
        sampled_lines = "".join(prefix_lines +
                                random.choices(poem_lines, k=num_lines_needed))

        prompt = f"{base_prompt}{sampled_lines}"
        message = [
            {
                "role": "user",
                "content": prompt,
            },
        ]
        prompt_formatted = tokenizer.apply_chat_template(
            message, add_generation_prompt=True, tokenize=False)
        prompt_len = len(tokenizer(prompt_formatted).input_ids)
        sampled_requests.append(
            (prompt, prompt_formatted, prompt_len, output_len, None))

    return sampled_requests


def sample_mmmu_pro_vision_requests(
    dataset,
    num_requests: int,
    tokenizer: PreTrainedTokenizerBase,
    fixed_output_len: Optional[int] = None,
) -> List[Tuple[str, str, int, Optional[Dict[str, Collection[str]]]]]:
    sampled_requests: List[Tuple[str, int, int, Dict[str,
                                                     Collection[str]]]] = []
    for data in dataset:
        if len(sampled_requests) == num_requests:
            break

        # MMMU-Pro vision direct prompt
        # Ref: https://github.com/MMMU-Benchmark/MMMU/blob/6ce42f4d8f70c1841c67867152648974415b5cac/mmmu-pro/prompts.yaml#L5
        prompt = (
            "Answer with the option letter from the given choices directly. "
            "The last line of your response should be of the following "
            "format: 'Answer: $LETTER' (without quotes) where LETTER is one of "
            "options.")

        prompt_token_ids = tokenizer(prompt).input_ids
        if fixed_output_len is None:
            # Default max output len is set to 128
            print("--hf-output-len is not provided. Using default value 128.")
            fixed_output_len = 128

        prompt_len = len(prompt_token_ids)
        output_len = fixed_output_len

        assert isinstance(
            data["image"],
            Image), ("Input image format must be `PIL.Image.Image`, "
                     f"given {type(data['image'])}.")
        image: Image = data["image"]
        image = image.convert("RGB")
        image_data = io.BytesIO()
        image.save(image_data, format='JPEG')
        image_base64 = base64.b64encode(image_data.getvalue()).decode("utf-8")
        mm_content = {
            "type": "image_url",
            "image_url": {
                "url": f"data:image/jpeg;base64,{image_base64}"
            },
        }

        sampled_requests.append((prompt, prompt_len, output_len, mm_content))

    return sampled_requests


def sample_hf_requests(
    dataset_path: str,
    dataset_subset: str,
    dataset_split: str,
    num_requests: int,
    tokenizer: PreTrainedTokenizerBase,
    random_seed: int,
    fixed_output_len: Optional[int] = None,
) -> List[Tuple[str, str, int, Optional[Dict[str, Collection[str]]]]]:

    # Special case for MMMU-Pro vision dataset
    if dataset_path == 'MMMU/MMMU_Pro' and dataset_subset == 'vision':
        assert dataset_split == "test"
        dataset = load_dataset(dataset_path,
                               name=dataset_subset,
                               split=dataset_split,
                               streaming=True)
        assert "image" in dataset.features, (
            "MMMU/MMMU_Pro vision dataset must have 'image' column.")
        filter_func = lambda x: isinstance(x["image"], Image)
        dataset = dataset.shuffle(seed=random_seed).filter(filter_func)
        return sample_mmmu_pro_vision_requests(dataset, num_requests,
                                               tokenizer, fixed_output_len)

    dataset = load_dataset(dataset_path,
                           name=dataset_subset,
                           split=dataset_split,
                           streaming=True)
    assert "conversations" in dataset.features, (
        "HF Dataset must have 'conversations' column.")
    filter_func = lambda x: len(x["conversations"]) >= 2
    filtered_dataset = dataset.shuffle(seed=random_seed).filter(filter_func)
    sampled_requests: List[Tuple[str, int, int, Dict[str,
                                                     Collection[str]]]] = []
    for data in filtered_dataset:
        if len(sampled_requests) == num_requests:
            break

        # Tokenize the prompts and completions.
        prompt = data["conversations"][0]["value"]
        prompt_token_ids = tokenizer(prompt).input_ids
        completion = data["conversations"][1]["value"]
        completion_token_ids = tokenizer(completion).input_ids
        prompt_len = len(prompt_token_ids)
        output_len = len(completion_token_ids
                         ) if fixed_output_len is None else fixed_output_len
        if fixed_output_len is None and (prompt_len < 4 or output_len < 4):
            # Prune too short sequences.
            continue
        if fixed_output_len is None and \
            (prompt_len > 1024 or prompt_len + output_len > 2048):
            # Prune too long sequences.
            continue

        if "image" in data and isinstance(data["image"], Image):
            image: Image = data["image"]
            image = image.convert("RGB")
            image_data = io.BytesIO()
            image.save(image_data, format='JPEG')
            image_base64 = base64.b64encode(
                image_data.getvalue()).decode("utf-8")
            mm_content = {
                "type": "image_url",
                "image_url": {
                    "url": f"data:image/jpeg;base64,{image_base64}"
                },
            }
        elif "image" in data and isinstance(data["image"], str):
            if (data["image"].startswith("http://") or \
                data["image"].startswith("file://")):
                image_url = data["image"]
            else:
                image_url = f"file://{data['image']}"

            mm_content = {
                "type": "image_url",
                "image_url": {
                    "url": image_url
                },
            }
        else:
            mm_content = None

        sampled_requests.append((prompt, prompt_len, output_len, mm_content))

    return sampled_requests


def sample_random_requests(
    prefix_len: int,
    input_len: int,
    output_len: int,
    num_prompts: int,
    range_ratio: float,
    tokenizer: PreTrainedTokenizerBase,
) -> List[Tuple[str, int, int]]:
    prefix_token_ids = np.random.randint(0,
                                         tokenizer.vocab_size,
                                         size=prefix_len).tolist()

    input_lens = np.random.randint(
        int(input_len * range_ratio),
        input_len + 1,
        size=num_prompts,
    )
    output_lens = np.random.randint(
        int(output_len * range_ratio),
        output_len + 1,
        size=num_prompts,
    )
    offsets = np.random.randint(0, tokenizer.vocab_size, size=num_prompts)
    input_requests = []
    for i in range(num_prompts):
        prompt = tokenizer.decode(prefix_token_ids +
                                  [(offsets[i] + i + j) % tokenizer.vocab_size
                                   for j in range(input_lens[i])])

        input_requests.append((prompt, int(prefix_len + input_lens[i]),
                               int(output_lens[i]), None))

    return input_requests


async def get_request(
    input_requests: List[Tuple[str, int, int]],
    request_rate: float,
    burstiness: float = 1.0,
) -> AsyncGenerator[Tuple[str, int, int], None]:
    """
    Asynchronously generates requests at a specified rate 
    with OPTIONAL burstiness.
    
    Args:
        input_requests: 
            A list of input requests, each represented as a tuple.
        request_rate: 
            The rate at which requests are generated (requests/s).
        burstiness (optional): 
            The burstiness factor of the request generation. 
            Only takes effect when request_rate is not inf.
            Default value is 1, which follows a Poisson process.
            Otherwise, the request intervals follow a gamma distribution.
            A lower burstiness value (0 < burstiness < 1) results 
            in more bursty requests, while a higher burstiness value 
            (burstiness > 1) results in a more uniform arrival of requests.
    """
    input_requests = iter(input_requests)

    # Calculate scale parameter theta to maintain the desired request_rate.
    assert burstiness > 0, (
        f"A positive burstiness factor is expected, but given {burstiness}.")
    theta = 1.0 / (request_rate * burstiness)

    for request in input_requests:
        yield request

        if request_rate == float("inf"):
            # If the request rate is infinity, then we don't need to wait.
            continue

        # Sample the request interval from the gamma distribution.
        # If burstiness is 1, it follows exponential distribution.
        interval = np.random.gamma(shape=burstiness, scale=theta)
        # The next request will be sent after the interval.
        await asyncio.sleep(interval)


def calculate_metrics(
    input_requests: List[Tuple[str, int, int]],
    outputs: List[RequestFuncOutput],
    dur_s: float,
    tokenizer: PreTrainedTokenizerBase,
    selected_percentile_metrics: List[str],
    selected_percentiles: List[float],
    gootput_config_dict: Dict[str, float],
) -> Tuple[BenchmarkMetrics, List[int]]:
    input_lens: List[int] = []
    actual_output_lens: List[int] = []
    actual_first_sentence_lens: List[int] = []
    total_input = 0
    completed = 0
    good_completed = 0
    itls: List[float] = []
    tpots: List[float] = []
    all_tpots: List[float] = []
    ttfts: List[float] = []
    ttfss: List[float] = []
    e2els: List[float] = []
    for i in range(len(outputs)):
        if outputs[i].success:
            # We use the tokenizer to count the number of output tokens for all
            # serving backends instead of looking at len(outputs[i].itl) since
            # multiple output tokens may be bundled together
            # Note : this may inflate the output token count slightly
            input_lens.append(input_requests[i][1])
            
            output_len = len(
                tokenizer(outputs[i].generated_text,
                          add_special_tokens=False).input_ids)
            actual_output_lens.append(output_len)
            
            first_sentence_len = len(
                tokenizer(outputs[i].first_sentence_text,
                          add_special_tokens=False).input_ids)
            actual_first_sentence_lens.append(first_sentence_len)
            
            total_input += input_requests[i][1]
            tpot = 0
            if output_len > 1:
                tpot = (outputs[i].latency - outputs[i].ttft) / (output_len -
                                                                 1)
                tpots.append(tpot)
            # Note: if output_len <= 1, we regard tpot as 0 for goodput
            all_tpots.append(tpot)
            itls += outputs[i].itl
            ttfts.append(outputs[i].ttft)
            ttfss.append(outputs[i].ttfs)
            e2els.append(outputs[i].latency)
            completed += 1
        else:
            input_lens.append(0)
            actual_output_lens.append(0)
            actual_first_sentence_lens.append(0)

    if gootput_config_dict:
        valid_metrics = []
        slo_values = []

        if "ttft" in gootput_config_dict:
            valid_metrics.append(ttfts)
            slo_values.append(gootput_config_dict["ttft"] /
                              MILLISECONDS_TO_SECONDS_CONVERSION)
        if "ttfs" in gootput_config_dict:
            valid_metrics.append(ttfss)
            slo_values.append(gootput_config_dict["ttfs"] /
                              MILLISECONDS_TO_SECONDS_CONVERSION)
        if "tpot" in gootput_config_dict:
            valid_metrics.append(all_tpots)
            slo_values.append(gootput_config_dict["tpot"] /
                              MILLISECONDS_TO_SECONDS_CONVERSION)
        if "e2el" in gootput_config_dict:
            valid_metrics.append(e2els)
            slo_values.append(gootput_config_dict["e2el"] /
                              MILLISECONDS_TO_SECONDS_CONVERSION)

        for req_metric in zip(*valid_metrics):
            is_good_req = all([s >= r for s, r in zip(slo_values, req_metric)])
            if is_good_req:
                good_completed += 1

    if completed == 0:
        warnings.warn(
            "All requests failed. This is likely due to a misconfiguration "
            "on the benchmark arguments.",
            stacklevel=2)
    metrics = BenchmarkMetrics(
        completed=completed,
        total_input=total_input,
        total_output=sum(actual_output_lens),
        request_throughput=completed / dur_s,
        request_goodput=good_completed / dur_s,
        output_throughput=sum(actual_output_lens) / dur_s,
        total_token_throughput=(total_input + sum(actual_output_lens)) / dur_s,
        mean_input_token_len=np.mean(input_lens or 0),
        median_input_token_len=np.std(input_lens or 0),
        std_input_token_len=np.median(input_lens or 0),
        percentiles_input_token_len=[(p, np.percentile(input_lens or 0, p))
                             for p in selected_percentiles], 
        mean_output_token_len=np.mean(actual_output_lens or 0),
        median_output_token_len=np.std(actual_output_lens or 0),
        std_output_token_len=np.median(actual_output_lens or 0),
        percentiles_output_token_len=[(p, np.percentile(actual_output_lens or 0, p))
                             for p in selected_percentiles], 
        mean_fs_token_len=np.mean(actual_first_sentence_lens or 0),
        median_fs_token_len=np.std(actual_first_sentence_lens or 0),
        std_fs_token_len=np.median(actual_first_sentence_lens or 0),
        percentiles_fs_token_len=[(p, np.percentile(actual_first_sentence_lens or 0, p))
                             for p in selected_percentiles], 
        mean_ttft_ms=np.mean(ttfts or 0) *
        1000,  # ttfts is empty if streaming is not supported by backend
        std_ttft_ms=np.std(ttfts or 0) * 1000,
        median_ttft_ms=np.median(ttfts or 0) * 1000,
        percentiles_ttft_ms=[(p, np.percentile(ttfts or 0, p) * 1000)
                             for p in selected_percentiles],
        mean_ttfs_ms=np.mean(ttfss or 0) *
        1000,  # ttfss is empty if streaming is not supported by backend
        std_ttfs_ms=np.std(ttfss or 0) * 1000,
        median_ttfs_ms=np.median(ttfss or 0) * 1000,
        percentiles_ttfs_ms=[(p, np.percentile(ttfss or 0, p) * 1000)
                             for p in selected_percentiles],
        mean_tpot_ms=np.mean(tpots or 0) * 1000,
        std_tpot_ms=np.std(tpots or 0) * 1000,
        median_tpot_ms=np.median(tpots or 0) * 1000,
        percentiles_tpot_ms=[(p, np.percentile(tpots or 0, p) * 1000)
                             for p in selected_percentiles],
        mean_itl_ms=np.mean(itls or 0) * 1000,
        std_itl_ms=np.std(itls or 0) * 1000,
        median_itl_ms=np.median(itls or 0) * 1000,
        percentiles_itl_ms=[(p, np.percentile(itls or 0, p) * 1000)
                            for p in selected_percentiles],
        mean_e2el_ms=np.mean(e2els or 0) * 1000,
        std_e2el_ms=np.std(e2els or 0) * 1000,
        median_e2el_ms=np.median(e2els or 0) * 1000,
        percentiles_e2el_ms=[(p, np.percentile(e2els or 0, p) * 1000)
                             for p in selected_percentiles],
    )

    return metrics, actual_output_lens, actual_first_sentence_lens


async def benchmark(
    backend: str,
    api_url: str,
    base_url: str,
    model_id: str,
    tokenizer: PreTrainedTokenizerBase,
    input_requests: List[Tuple[str, int, int]],
    logprobs: Optional[int],
    best_of: int,
    request_rate: float,
    burstiness: float,
    disable_tqdm: bool,
    profile: bool,
    selected_percentile_metrics: List[str],
    selected_percentiles: List[str],
    ignore_eos: bool,
    gootput_config_dict: Dict[str, float],
    max_concurrency: Optional[int],
):
    if backend in ASYNC_REQUEST_FUNCS:
        request_func = ASYNC_REQUEST_FUNCS[backend]
    else:
        raise ValueError(f"Unknown backend: {backend}")

    print("Starting initial single prompt test run...")
    test_prompt, test_prompt_len, test_output_len, test_mm_content = (
        input_requests[0])
    if backend != "openai-chat" and test_mm_content is not None:
        # multi-modal benchmark is only available on OpenAI Chat backend.
        raise ValueError(
            "Multi-modal content is only supported on 'openai-chat' backend.")
    test_input = RequestFuncInput(
        model=model_id,
        prompt=test_prompt,
        api_url=api_url,
        prompt_len=test_prompt_len,
        output_len=test_output_len,
        logprobs=logprobs,
        best_of=best_of,
        multi_modal_content=test_mm_content,
        ignore_eos=ignore_eos,
    )
    test_output = await request_func(request_func_input=test_input)
    if not test_output.success:
        raise ValueError(
            "Initial test run failed - Please make sure benchmark arguments "
            f"are correctly specified. Error: {test_output.error}")
    else:
        print("Initial test run completed. Starting main benchmark run...")

    if profile:
        print("Starting profiler...")
        profile_input = RequestFuncInput(model=model_id,
                                         prompt=test_prompt,
                                         api_url=base_url + "/start_profile",
                                         prompt_len=test_prompt_len,
                                         output_len=test_output_len,
                                         logprobs=logprobs,
                                         best_of=best_of,
                                         multi_modal_content=test_mm_content,
                                         ignore_eos=ignore_eos)
        profile_output = await request_func(request_func_input=profile_input)
        if profile_output.success:
            print("Profiler started")

    if burstiness == 1.0:
        distribution = "Poisson process"
    else:
        distribution = "Gamma distribution"

    print(f"Traffic request rate: {request_rate}")
    print(f"Burstiness factor: {burstiness} ({distribution})")
    print(f"Maximum request concurrency: {max_concurrency}")

    pbar = None if disable_tqdm else tqdm(total=len(input_requests))

    # This can be used once the minimum Python version is 3.10 or higher,
    # and it will simplify the code in limited_request_func.
    #    semaphore = (asyncio.Semaphore(max_concurrency)
    #                 if max_concurrency else contextlib.nullcontext())
    semaphore = (asyncio.Semaphore(max_concurrency)
                 if max_concurrency else None)

    async def limited_request_func(request_func_input, pbar):
        if semaphore is None:
            return await request_func(request_func_input=request_func_input,
                                      pbar=pbar)
        async with semaphore:
            return await request_func(request_func_input=request_func_input,
                                      pbar=pbar)

    benchmark_start_time = time.perf_counter()
    tasks: List[asyncio.Task] = []
    async for request in get_request(input_requests, request_rate, burstiness):
        prompt, prompt_len, output_len, mm_content = request
        request_func_input = RequestFuncInput(model=model_id,
                                              prompt=prompt,
                                              api_url=api_url,
                                              prompt_len=prompt_len,
                                              output_len=output_len,
                                              logprobs=logprobs,
                                              best_of=best_of,
                                              multi_modal_content=mm_content,
                                              ignore_eos=ignore_eos)
        tasks.append(
            asyncio.create_task(
                limited_request_func(request_func_input=request_func_input,
                                     pbar=pbar)))
    outputs: List[RequestFuncOutput] = await asyncio.gather(*tasks)

    if profile:
        print("Stopping profiler...")
        profile_input = RequestFuncInput(
            model=model_id,
            prompt=test_prompt,
            api_url=base_url + "/stop_profile",
            prompt_len=test_prompt_len,
            output_len=test_output_len,
            logprobs=logprobs,
            best_of=best_of,
        )
        profile_output = await request_func(request_func_input=profile_input)
        if profile_output.success:
            print("Profiler stopped")

    if pbar is not None:
        pbar.close()

    benchmark_duration = time.perf_counter() - benchmark_start_time

    metrics, actual_output_lens, first_sentence_lens = calculate_metrics(
        input_requests=input_requests,
        outputs=outputs,
        dur_s=benchmark_duration,
        tokenizer=tokenizer,
        selected_percentile_metrics=selected_percentile_metrics,
        selected_percentiles=selected_percentiles,
        gootput_config_dict=gootput_config_dict,
    )

    print("{s:{c}^{n}}".format(s=' Serving Benchmark Result ', n=50, c='='))
    print("{:<40} {:<10}".format("Successful requests:", metrics.completed))
    print("{:<40} {:<10.2f}".format("Benchmark duration (s):",
                                    benchmark_duration))
    print("{:<40} {:<10}".format("Total input tokens:", metrics.total_input))
    print("{:<40} {:<10}".format("Total generated tokens:",
                                 metrics.total_output))
    print("{:<40} {:<10.2f}".format("Request throughput (req/s):",
                                    metrics.request_throughput))
    if gootput_config_dict:
        print("{:<40} {:<10.2f}".format("Request goodput (req/s):",
                                        metrics.request_goodput))
    print("{:<40} {:<10.2f}".format("Output token throughput (tok/s):",
                                    metrics.output_throughput))
    print("{:<40} {:<10.2f}".format("Total Token throughput (tok/s):",
                                    metrics.total_token_throughput))

    result = {
        "duration": benchmark_duration,
        "completed": metrics.completed,
        "total_input_tokens": metrics.total_input,
        "total_output_tokens": metrics.total_output,
        "request_throughput": metrics.request_throughput,
        "request_goodput:":
        metrics.request_goodput if gootput_config_dict else None,
        "output_throughput": metrics.output_throughput,
        "total_token_throughput": metrics.total_token_throughput,
        "input_lens": [output.prompt_len for output in outputs],
        "output_lens": actual_output_lens,
        "first_sentence_lens": first_sentence_lens,
        "ttfts": [output.ttft for output in outputs],
        "ttfss": [output.ttfs for output in outputs],
        "itls": [output.itl for output in outputs],
        "generated_texts": [output.generated_text for output in outputs],
        "errors": [output.error for output in outputs],
    }

    def process_one_metric(
        # E.g., "ttft"
        metric_attribute_name: str,
        # E.g., "TTFT"
        metric_name: str,
        # E.g., "Time to First Token"
        metric_header: str,
        # E.g., "ms"
        metric_unit: str,
    ):
        # This function prints and adds statistics of the specified
        # metric.
        if metric_attribute_name not in selected_percentile_metrics:
            return
        print("{s:{c}^{n}}".format(s=metric_header, n=50, c='-'))
        print("{:<40} {:<10.2f}".format(
            f"Mean {metric_name} ({metric_unit}):",
            getattr(metrics, f"mean_{metric_attribute_name}_{metric_unit}")))
        print("{:<40} {:<10.2f}".format(
            f"Median {metric_name} ({metric_unit}):",
            getattr(metrics, f"median_{metric_attribute_name}_{metric_unit}")))
        result[f"mean_{metric_attribute_name}_{metric_unit}"] = getattr(
            metrics, f"mean_{metric_attribute_name}_{metric_unit}")
        result[f"median_{metric_attribute_name}_{metric_unit}"] = getattr(
            metrics, f"median_{metric_attribute_name}_{metric_unit}")
        result[f"std_{metric_attribute_name}_{metric_unit}"] = getattr(
            metrics, f"std_{metric_attribute_name}_{metric_unit}")
        for p, value in getattr(metrics,
                                f"percentiles_{metric_attribute_name}_{metric_unit}"):
            p_word = str(int(p)) if int(p) == p else str(p)
            print("{:<40} {:<10.2f}".format(f"P{p_word} {metric_name} ({metric_unit}):",
                                            value))
            result[f"p{p_word}_{metric_attribute_name}_{metric_unit}"] = value

    process_one_metric("input_token", "Input Tokens", "Stats for Input Tokens Len", "len")
    process_one_metric("output_token", "Output Tokens", "Stats for Output Tokens Len", "len")
    process_one_metric("fs_token", "First Sentence Tokens", "Stats for First Sentence Tokens Len", "len")
    process_one_metric("ttfs", "TTFS", "Time to First Sentence", "ms")
    process_one_metric("tpot", "TPOT",
                       "Time per Output Token (excl. 1st token)", "ms")
    process_one_metric("itl", "ITL", "Inter-token Latency", "ms")
    process_one_metric("e2el", "E2EL", "End-to-end Latency", "ms")

    print("=" * 50)

    return result


def check_goodput_args(args):
    # Check and parse goodput arguments
    gootput_config_dict = {}
    VALID_NAMES = ["ttft", "ttfs", "tpot", "e2el"]
    if args.goodput:
        gootput_config_dict = parse_goodput(args.goodput)
        for slo_name, slo_val in gootput_config_dict.items():
            if slo_name not in VALID_NAMES:
                raise ValueError(
                    f"Invalid metric name found, {slo_name}: {slo_val}. "
                    "The service level objective name should be one of "
                    f"{str(VALID_NAMES)}. ")
            if slo_val < 0:
                raise ValueError(
                    f"Invalid value found, {slo_name}: {slo_val}. "
                    "The service level objective value should be "
                    "non-negative.")
    return gootput_config_dict


def parse_goodput(slo_pairs):
    gootput_config_dict = {}
    try:
        for slo_pair in slo_pairs:
            slo_name, slo_val = slo_pair.split(":")
            gootput_config_dict[slo_name] = float(slo_val)
    except ValueError as err:
        raise argparse.ArgumentTypeError(
            "Invalid format found for service level objectives. "
            "Specify service level objectives for goodput as \"KEY:VALUE\" "
            "pairs, where the key is a metric name, and the value is a "
            "number in milliseconds.") from err
    return gootput_config_dict





from abc import ABC, abstractmethod


class Processor(ABC):
    r'''
    function to handle a single data
    return None if no data needs to return. Otherwise, return a list of transformed data
    It is useful when certain filtering logic is needed
    '''
    @abstractmethod
    def transform(self, item, is_last_sample=False):
        pass

    r'''
    function to handle a batch of data
    input:
        list_of_items -- data to be batch processed
        is_last_batch -- whether current batch is the last batch might be useful if the last batch's 
                         tranform logic is different from the normal batches.
    return:
        the transformed batch data. None if no data needs to return, otherwise, it should be a list of batch data
        since buffer gathering logic might exist, and it is possible to return multiple of batch data.
    '''
    @abstractmethod
    def batch_transform(self, list_of_items, is_last_batch=False):
        pass


class AncProcessor(Processor):
    r'''
    A processor that does nothing but return the input as is for both transform and batch transform function
    '''
    def transform(self, item, is_last_sample=False):
        return [item]

    # we wrap out data to a list to follow the return pattern in the base class
    def batch_transform(self, list_of_items, is_last_batch=False):
        return [list_of_items]


# AncComposerBuffer uses get_token_length_fn to get the token length of each sample.
# Set get_token_length_fn properly to make sample composing correct.
class AncComposerBuffer:
    def __init__(self, get_token_length_fn=None):
        self.buffer = []
        self.total_token_length = 0
        if get_token_length_fn is None:
            self.get_token_length_fn = lambda x: 1
        else:
            self.get_token_length_fn = get_token_length_fn
        self.delete_idx = set()
        self.cur_idx = 0

    def put(self, item):
        if item is not None:
            token_length = self.get_token_length_fn(item)
            self.buffer.append((token_length, item))
            self.total_token_length += token_length
        return self.total_token_length

    # return the first sample whose token length is smaller than the target length
    def get(self, target_length):
        for i in range(len(self.buffer)):
            idx = (i + self.cur_idx) % len(self.buffer)
            if idx in self.delete_idx:
                continue
            if self.buffer[idx][0] <= target_length:
                self.delete_idx.add(idx)
                self.total_token_length -= self.buffer[idx][0]
                self.cur_idx = (idx + 1) % len(self.buffer)
                return self.buffer[idx]
        return None
    
    def get_all(self):
        res = self.buffer
        self.buffer = []
        self.total_token_length = 0
        self.delete_idx = set()
        self.cur_ids = 0
        return res

    def flush(self):
        buffer = [self.buffer[i] for i in range(len(self.buffer)) if i not in self.delete_idx]
        self.buffer = buffer
        self.delete_idx = set()


class AncComposer:
    def __init__(self, max_seq_len, get_token_length_fn=None, ratio=2.0):
        self.max_seq_len = max_seq_len
        self.buffer = AncComposerBuffer(get_token_length_fn)
        # ratio here controls how many samples would be collected in the buffer
        # it might be difficult to compose enough samples to reach the max_seq_len when buffer size is small
        self.ratio = ratio

    def get_items(self, is_last_call=False):
        items = []
        if is_last_call:
            items = self.buffer.get_all()
            items = [i[1] for i in items if i is not None]
            return items
        cur_seq_length = self.max_seq_len
        while True:
            item = self.buffer.get(cur_seq_length)
            if item is None:
                break
            assert len(item) == 2
            items.append(item[1])
            cur_seq_length = cur_seq_length - item[0]
            if cur_seq_length == 0:
                break
        self.buffer.flush()
        return items

    def apply(self, sample, is_last_call=False):
        total_token_length = self.buffer.put(sample)
        res = []
        if total_token_length >= self.max_seq_len * self.ratio:
            items = self.get_items()
            res.append(items)
        if is_last_call:
            while total_token_length >= self.max_seq_len:
                items = self.get_items()
                res.append(items)
                total_token_length = self.buffer.put(None)
            if total_token_length > 0:
                items = self.get_items(is_last_call=True)
                res.append(items)
        return res if len(res) > 0 else None
    
    def __call__(self, samples, is_last_call=False):
        for sample in samples:
            res = self.apply(sample)
            if res is None:
                continue
            yield from res
        if is_last_call:
            res = self.apply(None, True)
            if res is not None:
                yield from res



from anc.data.anc_processor import Processor
from anc.data.processors.tokenizer import HFPretrainedTokenizer
import logging
import torch
from typing import List, Tuple


class FakeOpt:
    def __init__(self, hf_model_name, prompt_style, add_bos):
        self.hf_model_name = hf_model_name
        self.prompt_style = prompt_style
        self.prefix_template = ""
        self.delimiter = ""
        self.add_bos = add_bos


def setup_separate_prompt(opt):
    if opt.prompt_style == 'sunhao_old':
        opt.prefix_template = '### {role}:'
        opt.delimiter = '\n\n'
    # elif opt.prompt_style == 'chatml':
    #     opt.prefix_template = '<|im_start|> {role}\n'
    #     opt.delimiter = '<|im_end|>\n'
    elif opt.prompt_style == 'chatml2':
        opt.prefix_template = '<|im_start|>{role}\n'
        opt.delimiter = '<|im_end|>\n'
    elif opt.prompt_style == 'chatml_non_special': # used when <|im_end|> is not a special token
        opt.prefix_template = '<|im_start|> {role}\n'
        opt.delimiter = ' <|im_end|>\n'
    elif opt.prompt_style == 'llama3':
        opt.prefix_template = '<|start_header_id|>{role}<|end_header_id|>\n\n'
        opt.delimiter = '<|eot_id|>'
    else:
        # Use chatml2 by default
        opt.prefix_template = '<|im_start|>{role}\n'
        opt.delimiter = '<|im_end|>\n'

    opt.delimiter = opt.delimiter.replace('\\n', '\n')
    logging.info(f'Using dialog template:\n{opt.prefix_template}content{opt.delimiter}')


class SFTProcessor(Processor):
    def __init__(self, hf_model_name, prompt_style, max_seq_len, add_bos, batch_size, micro_batch_size, ceil_to_nearest=False):
        self.opt = FakeOpt(hf_model_name, prompt_style, add_bos)
        setup_separate_prompt(self.opt)
        self.tokenizer = HFPretrainedTokenizer(self.opt)
        self.max_seq_len = max_seq_len
        self.batch_size = batch_size
        self.micro_batch_size = micro_batch_size
        assert self.micro_batch_size == 1
        self.micro_batch = []
        self.pad_seq_length_to_mult = 16
        self.ceil_to_nearest = ceil_to_nearest

    def transform(self, item, is_last_sample=False):
        if not all(turn['role'] and turn['content'] and type(turn['role']) is str and type(turn['content']) is str for turn in item['messages']):
            return
        context = [(turn['role'].strip(), turn['content']) for turn in item['messages']]
        last_turn_mask_flag = item.get('last_turn_mask', False) or item.get('loss_on_last_turn', False)
        no_mask_flag = item.get('no_mask', False)
        context_vec, loss_mask = self._build_chitchat_prompt(
            context,
            self.opt.delimiter,
            loss_on_last_turn=last_turn_mask_flag,
            no_query_mask=no_mask_flag,
        )
        if self.opt.add_bos and context_vec[0] != self.tokenizer.start_token_id:
            assert type(self.tokenizer.start_token_id) is int
            context_vec.insert(0, self.tokenizer.start_token_id)
            loss_mask.insert(0, 0)
        # append one more token at the end of each sequence since nemo would truncate one
        # context_vec.append(self.tokenizer.end_token_id)
        # loss_mask.append(0)
        yield from [{"input_ids": context_vec, "loss_mask": loss_mask}]

    def _build_chitchat_prompt(
        self,
        context: List[Tuple[str, str]], 
        dialog_sep: str=None, 
        loss_on_last_turn=False, 
        no_query_mask=False, 
        add_generation_suffix=False, 
        **kwargs
    ):
        
        assert not (loss_on_last_turn and no_query_mask)
        if dialog_sep is None:
            dialog_sep: str = self.opt.delimiter

        def _encode_one(prefix: str, content: str, suffix: str, is_output: bool = False):
            if is_output or no_query_mask:
                encoded_role, encoded_content, encoded_suffix = (
                    self.tokenizer.txt2vec(prefix), 
                    self.tokenizer.txt2vec(content + suffix),
                    []
                )
            else:
                encoded_role, encoded_content, encoded_suffix = (
                    [],
                    self.tokenizer.txt2vec(prefix + content + suffix),
                    []
                )
            return encoded_role, encoded_content, encoded_suffix

        idxs_encoded = []
        idxs_mask = []

        prefix_template: str = self.opt.prefix_template
        # sep_ids = self.tokenizer.txt2vec(dialog_sep)

        for i, (role, content) in enumerate(context):
            suffix_sep = dialog_sep # if i < len(context) - 1 or add_generation_suffix else ""
                
            if role == 'assistant':
                # response turn
                encoded_role, encoded_content, encoded_suffix = _encode_one(prefix_template.format(role=role), content, suffix_sep, is_output=True)
                content_mask = [int(not loss_on_last_turn or i == len(context) - 1)] * len(encoded_content)
            else:
                if role not in ('user', 'human', 'system', 'document', 'memory', 'function', 'ipython', 'tool', 'tool_calls'):
                    logging.warn(f"Found unknown role={role}")
                encoded_role, encoded_content, encoded_suffix = _encode_one(prefix_template.format(role=role), content, suffix_sep)
                content_mask = [int(no_query_mask)] * len(encoded_content)

            idxs_encoded.extend(encoded_role + encoded_content + encoded_suffix)
            idxs_mask.extend([0] * len(encoded_role) + content_mask + [0] * len(encoded_suffix))

        if add_generation_suffix:
            idxs_encoded.extend(self.tokenizer.txt2vec(prefix_template.format(role='assistant')))
            idxs_mask.extend([0] * (len(idxs_encoded) - len(idxs_mask)))

        assert len(idxs_encoded) == len(idxs_mask)
        return idxs_encoded, idxs_mask
    
    def get_token_length_fn(self, item):
        return len(item['input_ids'])

    def _compose_one_sequence(self, items):
        input_ids = []
        loss_mask = []
        labels = []
        position_ids = []
        cu_seqlens = []
        cur_seqlen = 0
        if not isinstance(items, list):
            items = [items]
        for item in items:
            input_ids += item['input_ids']
            labels += item['input_ids'][1:] + [self.tokenizer.end_token_id]
            loss_mask += item['loss_mask'][1:] + [0]
            cu_seqlens.append(cur_seqlen)
            cur_seqlen += len(item['input_ids'])
            position_ids += list(range(len(item['input_ids'])))
        cu_seqlens.append(cur_seqlen)
        token_count = len(input_ids)

        res = {}
        res['tokens'] = torch.LongTensor(input_ids)
        res['labels'] = torch.LongTensor(labels)
        res['loss_mask'] = torch.LongTensor(loss_mask)
        res['position_ids'] = torch.LongTensor(position_ids)
        res['token_count'] = token_count
        res['cu_seqlens'] = cu_seqlens
        return res

    def _collate_item(self, item, max_length, pad_id):
        item = torch.nn.functional.pad(item, (0, max_length - item.shape[0]), value=pad_id)
        return item

    def create_micro_batch(self, items):
        res = {}
        max_token_count = max([i['token_count'] for i in items])
        if self.ceil_to_nearest:
            max_token_count = self._ceil_to_nearest(max_token_count, self.pad_seq_length_to_mult)
        for k in items[0]:
            if k in {'tokens', 'labels'}:
                res[k] = torch.stack([self._collate_item(item[k], max_token_count, self.tokenizer.end_token_id) for item in items])
            elif k in {'loss_mask', 'position_ids'}:
                res[k] = torch.stack([self._collate_item(item[k], max_token_count, 0) for item in items])
            elif k == 'token_count':
                res[k] = [item['token_count'] for item in items]
            elif k == 'cu_seqlens':
                for item in items:
                    if item[k][-1] < max_token_count:
                        item[k].append(max_token_count)
                    item[k] = torch.IntTensor(item[k])
                max_seq_count = max(len(item[k]) for item in items)
                res[k] = torch.stack([self._collate_item(item[k], max_seq_count + 1, -1) for item in items])
        res['attention_mask'] = torch.LongTensor([1] * len(items))
        res['cu_seqlens_argmin'] = torch.argmin(res['cu_seqlens'], dim=1, keepdim=True)
        seqlens = res['cu_seqlens'][:, 1:] - res['cu_seqlens'][:, :-1]
        max_seqlen, _ = seqlens.max(dim=1, keepdim=True)
        res['max_seqlen'] = max_seqlen
        # create cu_seqlens_unpadded, since we didn't pad any token between sequences during the sequence composing
        # we simply clone the cu_seqlens and cu_seqlens_argmin here
        res['cu_seqlens_unpadded'] = res['cu_seqlens'].clone()
        res['cu_seqlens_unpadded_argmin'] = res['cu_seqlens_argmin'].clone()
        return res

    def _ceil_to_nearest(self, n, m):
        return (n + m - 1) // m * m
    
    def batch_transform(self, list_of_items, is_last_batch=False):
        for items in list_of_items:
            one_sequence = self._compose_one_sequence(items)
            self.micro_batch.append(one_sequence)
            if len(self.micro_batch) == self.batch_size:
                one_batch = self.create_micro_batch(self.micro_batch)
                self.micro_batch = []
                yield one_batch
        if is_last_batch:
            if self.micro_batch:
                one_batch = self.create_micro_batch(self.micro_batch)
                self.micro_batch = []
                yield one_batch

    def __getstate__(self):
        return self.__dict__
    
    def __setstate__(self, states):
        self.__dict__.update(states)


class PPOProcessor:
    def __init__(self, cfg, tokenizer, collate_fn):
        self.tokenizer = tokenizer
        self.add_eod = cfg.model.data.get("append_eod", False)
        seq_length = cfg.model.data.seq_length
        if "length_params" in cfg.model.ppo:
            max_sample_length = seq_length - cfg.model.ppo.length_params.max_length
        else:
            max_sample_length = seq_length // 2
        self.max_sample_length = max_sample_length
        self.collate_fn = collate_fn

    def transform(self, sample, is_last_sample=False):
        text = sample["text"]
        text_ids = self.tokenizer.text_to_ids(text)
        if len(text_ids) > 0 and self.add_eod:
            text_ids.append(self.tokenizer.eos_id)
        if len(text_ids) > self.max_sample_length:
            return
        sample_tensor = torch.tensor(text_ids, dtype=torch.int64)
        output = {
            "text": sample_tensor,
            "length": sample_tensor.shape[0],
            "loss_multiplier": True,
        }
        return [output]
    
    def batch_transform(self, items, is_last_batch=False):
        return [self.collate_fn(items)]
